<?php

namespace Drupal\admin_ui_only;

use Drupal\Core\Routing\RouteBuildEvent;
use Drupal\Core\Routing\RouteObjectInterface;
use Drupal\Core\Routing\RoutingEvents;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Event\RequestEvent;
use Symfony\Component\HttpKernel\Event\ResponseEvent;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Symfony\Component\HttpKernel\KernelEvents;

/**
 * Prevents html access to non admin pages.
 */
class EventSubscriber implements EventSubscriberInterface {

  /**
   * A list of routes that should be admin routes.
   */
  private const ADMIN_ROUTES = [
    'entity.user.cancel_form',
    'entity.user.canonical',
    'entity.user.contact_form',
    'entity.user.edit_form',
    'system.batch_page.html',
    'system.401',
    'system.403',
    'system.404',
    'system.4xx',
    'user.edit',
    'user.login',
    'user.logout',
    'user.page',
    'user.cancel_confirm',
    'user.reset.login',
    'user.reset',
    'user.reset.form',
    'user.well-known.change_password',
    'user.register',
  ];

  /**
   * Triggers a 404 for applicable routes on request.
   *
   * @param \Symfony\Component\HttpKernel\Event\RequestEvent $event
   *   The event to process.
   */
  public function onKernelRequest(RequestEvent $event): void {
    if ($event->isMainRequest()) {
      $request = $event->getRequest();
      // We're only interested in GET, HEAD, OPTIONS and TRACE methods. This
      // allows contextual links to be rendered by POSTing to contextual/render.
      if ($request->isMethodSafe() && $this->deny($request)) {
        // @todo make the exception type configurable.
        throw new NotFoundHttpException();
      }
    }
  }

  /**
   * Triggers a 404 for applicable routes on response.
   *
   * This allows us to reject POST requests to paths like node/1.
   *
   * @param \Symfony\Component\HttpKernel\Event\ResponseEvent $event
   *   The event to process.
   */
  public function onKernelRespond(ResponseEvent $event): void {
    if ($event->isMainRequest()) {
      $response = $event->getResponse();
      if (
        $response->getStatusCode() === 200 &&
        // Checking the content type allows JSON responses through.
        str_contains($response->headers->get('Content-Type', ''), 'text/html') &&
        $this->deny($event->getRequest())
      ) {
        // @todo make the exception type configurable.
        throw new NotFoundHttpException();
      }
    }
  }

  /**
   * Determines if the request should be denied.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The request.
   *
   * @return bool
   *   TRUE if the request should be denied and FALSE if should be allowed.
   */
  private function deny(Request $request): bool {
    if ($request->getRequestFormat() !== 'html') {
      return FALSE;
    }
    /** @var \Symfony\Component\Routing\Route|null $route_object */
    $route_object = $request->attributes->get(RouteObjectInterface::ROUTE_OBJECT);
    if ($route_object && $route_object->getOption('_admin_route')) {
      // We are not interested in admin routes.
      return FALSE;
    }
    // Give access to front page. The front page should be set to user/login for
    // the best UX.
    return $request->getPathInfo() !== '/';
  }

  /**
   * Alters existing routes for a specific collection.
   *
   * @param \Drupal\Core\Routing\RouteBuildEvent $event
   *   The route build event.
   */
  public function onAlterRoutes(RouteBuildEvent $event): void {
    $collection = $event->getRouteCollection();
    // @todo Allow users to add more routes to this list.
    foreach (self::ADMIN_ROUTES as $route_name) {
      $route = $collection->get($route_name);
      $route?->setOption('_admin_route', TRUE);
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array {
    $events[KernelEvents::REQUEST][] = ['onKernelRequest', 30];
    $events[KernelEvents::RESPONSE][] = ['onKernelRespond'];
    $events[RoutingEvents::ALTER][] = ['onAlterRoutes', 30];
    return $events;
  }

}
